@@ -1,6 +1,6 @@
 use regex::Regex;
 use std::io::prelude::*;
-use std::sync::atomic::{AtomicI64, Ordering};
+use std::sync::atomic::{AtomicUsize, Ordering};
 use std::time::Duration;
 use std::{
     cmp::Reverse,
@@ -31,9 +31,10 @@ use rand::Rng;
 use rlimit::{setrlimit, Resource};
 use serde::{Deserialize, Serialize};
 use serde_json::Value;
-use sysinfo::{RefreshKind, System, SystemExt};
 use threadpool::ThreadPool;
 
+const MAX_PIXZ: usize = 2;
+
 #[derive(Serialize, Deserialize)]
 pub struct Message {
     /// The exchange name, unique for each exchage
@@ -48,21 +49,22 @@ pub struct Message {
     pub json: String,
 }
 
-fn get_day(unix_timestamp: i64) -> String {
-    let naive = NaiveDateTime::from_timestamp(unix_timestamp, 0);
+fn get_day(timestamp_millis: i64) -> String {
+    let naive = NaiveDateTime::from_timestamp(timestamp_millis / 1000, 0);
     let datetime: DateTime<Utc> = DateTime::from_utc(naive, Utc);
     datetime.format("%Y-%m-%d").to_string()
 }
 
+fn get_hour(timestamp_millis: i64) -> String {
+    let naive = NaiveDateTime::from_timestamp(timestamp_millis / 1000, 0);
+    let datetime: DateTime<Utc> = DateTime::from_utc(naive, Utc);
+    datetime.format("%Y-%m-%d-%H").to_string()
+}
+
 // Output to a raw file and a parsed file.
 #[derive(Clone)]
 #[allow(clippy::type_complexity)]
-struct Output(
-    Arc<(
-        Mutex<Box<dyn std::io::Write + Send>>,
-        Mutex<Box<dyn std::io::Write + Send>>,
-    )>,
-);
+struct Output(Arc<Mutex<Box<dyn std::io::Write + Send>>>);
 
 fn is_blocked_market(market_type: MarketType) -> bool {
     market_type == MarketType::QuantoFuture
@@ -127,8 +129,11 @@ where
                 };
                 if is_new {
                     if let Some(symbol) = extract_symbol(exchange, market_type, &msg.json) {
-                        let output = {
-                            if !split_files.contains_key(&symbol) {
+                        // raw
+                        if day == get_day(msg.received_at as i64) {
+                            let hour = get_hour(msg.received_at as i64);
+                            let key = format!("raw.{}.{}", symbol, hour);
+                            if !split_files.contains_key(&key) {
                                 let buf_writer_raw = {
                                     let output_file_name = format!(
                                         "{}.{}.{}.{}.{}.json.gz",
@@ -136,7 +141,7 @@ where
                                         market_type_str,
                                         msg_type_str,
                                         re.replace_all(&symbol, "_"),
-                                        day
+                                        hour
                                     );
                                     let f_out = std::fs::OpenOptions::new()
                                         .create(true)
@@ -152,7 +157,22 @@ where
                                         Compression::default(),
                                     ))
                                 };
-
+                                split_files.insert(
+                                    key.clone(),
+                                    Output(Arc::new(Mutex::new(Box::new(buf_writer_raw)))),
+                                );
+                            }
+                            let output = {
+                                let entry = split_files.get(&key).unwrap();
+                                entry.value().clone()
+                            };
+                            writeln!(output.0.lock().unwrap(), "{}", line).unwrap();
+                        }
+                        // parsed
+                        let write_parsed = |json: String, timestamp: i64| {
+                            let hour = get_hour(timestamp);
+                            let key = format!("parsed.{}.{}", symbol, hour);
+                            if !split_files.contains_key(&key) {
                                 let buf_writer_parsed = {
                                     let pair =
                                         crypto_pair::normalize_pair(&symbol, exchange).unwrap();
@@ -164,7 +184,7 @@ where
                                         msg_type_str,
                                         re.replace_all(&pair, "_"),
                                         re.replace_all(&symbol, "_"),
-                                        day
+                                        hour
                                     );
                                     let f_out = std::fs::OpenOptions::new()
                                         .create(true)
@@ -181,20 +201,17 @@ where
                                     ))
                                 };
                                 split_files.insert(
-                                    symbol.clone(),
-                                    Output(Arc::new((
-                                        Mutex::new(Box::new(buf_writer_raw)),
-                                        Mutex::new(Box::new(buf_writer_parsed)),
-                                    ))),
+                                    key.clone(),
+                                    Output(Arc::new(Mutex::new(Box::new(buf_writer_parsed)))),
                                 );
                             }
-                            let entry = split_files.get(&symbol).unwrap();
-                            entry.value().clone()
+                            let output = {
+                                let entry = split_files.get(&key).unwrap();
+                                entry.value().clone()
+                            };
+                            writeln!(output.0.lock().unwrap(), "{}", json).unwrap();
                         };
-                        // raw
-                        if day == get_day((msg.received_at / 1000_u64) as i64) {
-                            writeln!(output.0 .0.lock().unwrap(), "{}", line).unwrap();
-                        }
+
                         match msg.msg_type {
                             MessageType::L2Event => {
                                 // Skip unsupported markets
@@ -206,13 +223,11 @@ where
                                         Some(msg.received_at as i64),
                                     ) {
                                         for message in messages {
-                                            if get_day(message.timestamp / 1000) == day {
-                                                writeln!(
-                                                    output.0 .1.lock().unwrap(),
-                                                    "{}",
-                                                    serde_json::to_string(&message).unwrap()
-                                                )
-                                                .unwrap();
+                                            if get_day(message.timestamp) == day {
+                                                write_parsed(
+                                                    serde_json::to_string(&message).unwrap(),
+                                                    message.timestamp,
+                                                );
                                             }
                                         }
                                     }
@@ -223,13 +238,11 @@ where
                                     parse_trade(&msg.exchange, msg.market_type, &msg.json)
                                 {
                                     for message in messages {
-                                        if get_day(message.timestamp / 1000) == day {
-                                            writeln!(
-                                                output.0 .1.lock().unwrap(),
-                                                "{}",
-                                                serde_json::to_string(&message).unwrap()
-                                            )
-                                            .unwrap();
+                                        if get_day(message.timestamp) == day {
+                                            write_parsed(
+                                                serde_json::to_string(&message).unwrap(),
+                                                message.timestamp,
+                                            );
                                         }
                                     }
                                 }
@@ -253,37 +266,18 @@ where
     (error_lines, total_lines)
 }
 
-fn sort_file<P>(
-    input_file: P,
-    output_file: P,
-    use_pixz: bool,
-    available_memory: Arc<AtomicI64>,
-) -> (i64, i64)
+fn sort_file<P>(input_file: P, writer: &mut dyn std::io::Write) -> (i64, i64)
 where
     P: AsRef<Path>,
 {
     assert!(input_file.as_ref().to_str().unwrap().ends_with(".json.gz"));
-    assert!(output_file.as_ref().to_str().unwrap().ends_with(".json.xz"));
     if !input_file.as_ref().exists() {
         panic!("{:?} does not exist", input_file.as_ref().display());
     }
-
-    let estimated_memory_usage = {
-        let filesize = std::fs::metadata(input_file.as_ref()).unwrap().len();
-        (filesize * 5) as i64
+    let buf_reader = {
+        let f_in = std::fs::File::open(&input_file).unwrap();
+        std::io::BufReader::new(GzDecoder::new(f_in))
     };
-    {
-        // Make this thread sleep if memory is not enough, this section is optional
-        let mut rng = rand::thread_rng();
-        while available_memory.load(Ordering::SeqCst) < estimated_memory_usage {
-            let millis = rng.gen_range(1000_u64..5000_u64);
-            debug!("Available memory {} is less than estimated memory {}, sleeping for {} milliseconds", available_memory.load(Ordering::SeqCst), estimated_memory_usage, millis);
-            std::thread::sleep(Duration::from_millis(millis));
-        }
-    }
-
-    let f_in = std::fs::File::open(&input_file).unwrap();
-    let buf_reader = std::io::BufReader::new(GzDecoder::new(f_in));
     let mut total_lines = 0;
     let mut error_lines = 0;
     let mut lines: Vec<(i64, String)> = Vec::new();
@@ -311,61 +305,132 @@ where
             error_lines += 1;
         }
     }
+    std::fs::remove_file(input_file.as_ref()).unwrap();
     if error_lines == 0 {
         lines.sort_by_key(|x| x.0); // sort by timestamp
 
-        if !use_pixz {
-            let mut writer = {
-                let f_out = std::fs::OpenOptions::new()
-                    .create(true)
-                    .write(true)
-                    .truncate(true)
-                    .open(output_file)
-                    .unwrap();
-                let e = xz2::write::XzEncoder::new(f_out, 9);
-                std::io::BufWriter::new(e)
-            };
-            for line in lines {
-                writeln!(writer, "{}", line.1).unwrap();
+        for line in lines {
+            writeln!(writer, "{}", line.1).unwrap();
+        }
+        writer.flush().unwrap();
+    } else {
+        error!(
+            "Found {} malformed lines out of total {} lines in file {}",
+            error_lines,
+            total_lines,
+            input_file.as_ref().display()
+        );
+    }
+    (error_lines, total_lines)
+}
+
+// Use pixz if use_pixz is true, and semaphore allows only two pixz processes
+fn sort_files<P>(
+    mut hourly_files: Vec<P>,
+    output_file: P,
+    use_pixz: bool,
+    semaphore: Arc<AtomicUsize>,
+) -> (i64, i64)
+where
+    P: AsRef<Path>,
+{
+    for input_file in hourly_files.iter() {
+        assert!(input_file.as_ref().to_str().unwrap().ends_with(".json.gz"));
+        if !input_file.as_ref().exists() {
+            panic!("{:?} does not exist", input_file.as_ref().display());
+        }
+    }
+    if hourly_files.len() < 24 {
+        warn!(
+            "There are only {} files for {}",
+            hourly_files.len(),
+            output_file.as_ref().display()
+        );
+    }
+    hourly_files.sort_by(|a, b| a.as_ref().cmp(b.as_ref()));
+    assert!(output_file.as_ref().to_str().unwrap().ends_with(".json.xz"));
+
+    let mut writer: Box<dyn std::io::Write> = if !use_pixz {
+        let f_out = std::fs::OpenOptions::new()
+            .create(true)
+            .write(true)
+            .truncate(true)
+            .open(output_file.as_ref())
+            .unwrap();
+        let e = xz2::write::XzEncoder::new(f_out, 6);
+        Box::new(std::io::BufWriter::new(e))
+    } else {
+        let json_file = {
+            let output_dir = output_file.as_ref().parent().unwrap().to_path_buf();
+            let filename = output_file.as_ref().file_name().unwrap().to_str().unwrap();
+            output_dir.join(&filename[..filename.len() - 3])
+        };
+        let f_out = std::fs::OpenOptions::new()
+            .create(true)
+            .write(true)
+            .truncate(true)
+            .open(json_file.as_path())
+            .unwrap();
+        Box::new(std::io::BufWriter::new(f_out))
+    };
+
+    let mut total_lines = 0;
+    let mut error_lines = 0;
+    for input_file in hourly_files.iter() {
+        let (e, t) = sort_file(input_file, writer.as_mut());
+        total_lines += t;
+        error_lines += e;
+    }
+    drop(writer);
+    if error_lines == 0 {
+        if use_pixz {
+            {
+                // Wait for the semaphore which allows only two pixz processes
+                let mut rng = rand::thread_rng();
+                while semaphore.load(Ordering::SeqCst) == 0 {
+                    let millis = rng.gen_range(1000_u64..5000_u64);
+                    debug!("Waiting for semaphore");
+                    std::thread::sleep(Duration::from_millis(millis));
+                }
+                semaphore.fetch_sub(1_usize, Ordering::SeqCst);
             }
-            writer.flush().unwrap();
-        } else {
             let json_file = {
                 let output_dir = output_file.as_ref().parent().unwrap().to_path_buf();
                 let filename = output_file.as_ref().file_name().unwrap().to_str().unwrap();
                 output_dir.join(&filename[..filename.len() - 3])
             };
-            let mut writer = {
-                let f_out = std::fs::OpenOptions::new()
-                    .create(true)
-                    .write(true)
-                    .truncate(true)
-                    .open(json_file.as_path())
-                    .unwrap();
-                std::io::BufWriter::new(f_out)
-            };
-            for line in lines {
-                writeln!(writer, "{}", line.1).unwrap();
-            }
-            writer.flush().unwrap();
-            drop(writer);
             match std::process::Command::new("pixz")
-                .args(["-9", json_file.as_path().to_str().unwrap()])
+                .args(["-6", json_file.as_path().to_str().unwrap()])
                 .output()
             {
                 Ok(output) => {
-                    if output.status.success() {
-                        debug!("{}", String::from_utf8_lossy(&output.stdout));
-                    } else {
+                    if !output.status.success() {
                         panic!("{}", String::from_utf8_lossy(&output.stderr));
                     }
                 }
                 Err(err) => panic!("{}", err),
             }
+            semaphore.fetch_add(1_usize, Ordering::SeqCst);
+        }
+    } else {
+        error!(
+            "Found {} malformed lines out of total {} total lines for {}",
+            error_lines,
+            total_lines,
+            output_file.as_ref().display()
+        );
+        // cleanup
+        if use_pixz {
+            let json_file = {
+                let output_dir = output_file.as_ref().parent().unwrap().to_path_buf();
+                let filename = output_file.as_ref().file_name().unwrap().to_str().unwrap();
+                output_dir.join(&filename[..filename.len() - 3])
+            };
+            std::fs::remove_file(json_file.as_path()).unwrap();
+        } else {
+            std::fs::remove_file(output_file.as_ref()).unwrap();
         }
-        std::fs::remove_file(input_file).unwrap();
     }
-    available_memory.fetch_add(estimated_memory_usage, Ordering::SeqCst);
     (error_lines, total_lines)
 }
 
@@ -479,8 +544,7 @@ fn process_files_of_day(
         }
         for entry in split_files.iter() {
             let output = entry.value();
-            output.0 .0.lock().unwrap().flush().unwrap();
-            output.0 .1.lock().unwrap().flush().unwrap();
+            output.0.lock().unwrap().flush().unwrap();
         }
         let error_ratio = (error_lines as f64) / (total_lines as f64);
         if error_ratio > 0.01 {
@@ -497,14 +561,14 @@ fn process_files_of_day(
             );
             return false;
         } else {
-            info!("Finished split {} {} {} {}, dropped {} malformed lines out of {} lines, time elapsed {} seconds", exchange, market_type, msg_type, day, error_lines, total_lines, start_timstamp.elapsed().as_secs());
+            info!("Finished split {} {} {} {}, dropped {} malformed lines out of total {} lines, time elapsed {} seconds", exchange, market_type, msg_type, day, error_lines, total_lines, start_timstamp.elapsed().as_secs());
         }
     }
 
     // sort by timestamp
     {
         let glob_pattern = format!(
-            "{}/**/{}.{}.{}.*.{}.json.gz",
+            "{}/**/{}.{}.{}.*.{}-??.json.gz",
             output_dir_raw, exchange, market_type, msg_type, day
         );
         let paths_raw: Vec<PathBuf> = glob(&glob_pattern)
@@ -517,7 +581,7 @@ fn process_files_of_day(
         }
 
         let glob_pattern = format!(
-            "{}/**/{}.{}.{}.*.{}.json.gz",
+            "{}/**/{}.{}.{}.*.{}-??.json.gz",
             output_dir_parsed, exchange, market_type, msg_type, day
         );
         let paths_parsed: Vec<PathBuf> = glob(&glob_pattern)
@@ -525,7 +589,7 @@ fn process_files_of_day(
             .filter_map(Result::ok)
             .collect();
 
-        let mut paths = if is_blocked_market(market_type) {
+        let paths = if is_blocked_market(market_type) {
             for path in paths_parsed {
                 std::fs::remove_file(path).unwrap();
             }
@@ -537,6 +601,28 @@ fn process_files_of_day(
                 .collect()
         };
 
+        let paths_by_day = {
+            // group by day
+            let mut groups: HashMap<String, Vec<PathBuf>> = HashMap::new();
+            for path in paths {
+                let file_name = path.as_path().file_name().unwrap().to_str().unwrap();
+                let key = &file_name[0..(file_name.len() - "-??.json.gz".len())];
+                if !groups.contains_key(key) {
+                    groups.insert(key.to_string(), vec![]);
+                }
+                groups.get_mut(key).unwrap().push(path);
+            }
+            let mut groups: Vec<Vec<PathBuf>> = groups.values().cloned().collect();
+            // Smaller groups get processed first
+            groups.sort_by_cached_key(|group| {
+                group
+                    .iter()
+                    .map(|file| std::fs::metadata(file).unwrap().len())
+                    .sum::<u64>()
+            });
+            groups
+        };
+
         info!(
             "Started sort {} {} {} {}",
             exchange, market_type, msg_type, day
@@ -544,44 +630,31 @@ fn process_files_of_day(
         #[allow(clippy::type_complexity)]
         let (tx, rx): (Sender<(i64, i64)>, Receiver<(i64, i64)>) = mpsc::channel();
         let start_timstamp = Instant::now();
-        // Smaller files get processed first
-        paths.sort_by_cached_key(|path| std::fs::metadata(path).unwrap().len());
-        let percentile_90 = ((paths.len() as f64) * 0.9) as usize;
+        let percentile_90 = ((paths_by_day.len() as f64) * 0.9) as usize;
         let pixz_exists = Path::new("/usr/bin/pixz").exists();
-        let available_memory = {
-            let mut sys = System::new_with_specifics(RefreshKind::new().with_memory());
-            sys.refresh_memory();
-            let available_memory = sys.available_memory() * 1024;
-            Arc::new(AtomicI64::new(available_memory as i64))
-        };
-        for (index, input_file) in paths.into_iter().enumerate() {
-            {
-                let estimated_memory_usage = {
-                    let filesize = std::fs::metadata(input_file.as_path()).unwrap().len();
-                    (filesize * 5) as i64
-                };
-                available_memory.fetch_sub(estimated_memory_usage, Ordering::SeqCst);
-            }
-            let file_name = input_file.as_path().file_name().unwrap();
+        let semaphore = Arc::new(AtomicUsize::new(MAX_PIXZ));
+        for (index, input_files) in paths_by_day.into_iter().enumerate() {
+            let file_name = input_files[0]
+                .as_path()
+                .file_name()
+                .unwrap()
+                .to_str()
+                .unwrap();
             let output_file_name = format!(
                 "{}.json.xz",
-                file_name
-                    .to_str()
-                    .unwrap()
-                    .strip_suffix(".json.gz")
-                    .unwrap()
+                &file_name[0..(file_name.len() - "-??.json.gz".len())]
             );
-            let output_file = Path::new(input_file.parent().unwrap()).join(output_file_name);
+            let output_file = Path::new(input_files[0].parent().unwrap()).join(output_file_name);
             let tx_clone = tx.clone();
-            let available_memory_clone = available_memory.clone();
+            let semaphore_clone = semaphore.clone();
             if pixz_exists && index >= percentile_90 {
                 thread_pool.execute(move || {
-                    let t = sort_file(input_file, output_file, true, available_memory_clone);
+                    let t = sort_files(input_files, output_file, true, semaphore_clone);
                     tx_clone.send(t).unwrap();
                 });
             } else {
                 thread_pool.execute(move || {
-                    let t = sort_file(input_file, output_file, false, available_memory_clone);
+                    let t = sort_files(input_files, output_file, false, semaphore_clone);
                     tx_clone.send(t).unwrap();
                 });
             }
@@ -606,7 +679,7 @@ fn process_files_of_day(
             true
         } else {
             error!(
-                "Failed to sort {} {} {} {}, found {} malformed lines out of {} lines, time elapsed {} seconds",
+                "Failed to sort {} {} {} {}, found {} malformed lines out of total {} lines, time elapsed {} seconds",
                 exchange,
                 market_type,
                 msg_type,
@@ -621,7 +694,7 @@ fn process_files_of_day(
 
 fn main() {
     env_logger::init();
-    assert!(setrlimit(Resource::NOFILE, 4096, 4096).is_ok());
+    assert!(setrlimit(Resource::NOFILE, 131072, 131072).is_ok());
 
     let args: Vec<String> = env::args().collect();
     if args.len() != 8 {
